from collections import namedtuple
from statistics import mean
from typing import NamedTuple
from uuid import uuid4
from pathlib import Path

from programlib import language_

def correctness(expected_outputs, outputs):
    assert expected_outputs, 'expected_outputs is empty. Cannot test program correctness.'

    outputs = outputs[:len(expected_outputs)]

    if outputs:
        return mean(int(expected_line == actual_line) 
                    for expected_line, actual_line 
                    in zip(expected_outputs, outputs))
    else:
        return 0

TestRun = namedtuple('TestRun', ['input_lines', 'expected_output_lines', 
                                 'output_lines', 'correctness'])

class Program():
    """
    Program object: represents a runnable program in some programming language
    """

    def __init__(self, source, name=None, language='C++', 
                       workdir=Path(__file__).parent / 'programs'):
        self.language = language_(language)

        self.workdir = workdir
        self.name = name or str(uuid4())
        
        self.language.write_source(self.workdir, self.name, source)
        self.stdout, self.stderr = self.language.build(self.workdir, self.name)

        # Please send your opinions on 'not not' vs 'bool()' to dont@vadim.me
        self.compile_error = not not self.stderr

    def __lt__(self, other):
        return self.name < other.name

    def read(self):
        """Obtain the source code of the program"""

        return self.language.read_source(self.workdir, self.name)

    def run(self, input_lines=[], force=True):
        """
        Run the program and capture its output

        If 'input_lines' are specified, they will be sent to stdin.
        After the program has finished, `run` will check for errors on stderr
        and, in case there aren't any, return a list of output lines

        Use 'force=True' to skip the error check.
        Raw stdout and stderr data will always be stored in `program.stdout` and `program.stderr` attributes
        """
        
        self.stdout, self.stderr = self.language.run(self.workdir, self.name, input_lines)
        assert force or not self.stderr, self.stderr
        return self.stdout.splitlines()

    def test(self, test_cases, force=True, cache=True):
        """
        Test the program against a list of input output pairs

        If 'force=True', program outputs will be checked even if there are errors on stderr
        If 'cache=True', all the test logs will be stored in `program.test_runs` attribute
        """

        if self.compile_error:
            self.score = 0
        else:
            self.test_runs = []
            for input_lines, expected_output_lines in test_cases:
                try:
                    output_lines = self.run(input_lines, force=force)
                    test_run = TestRun(input_lines, expected_output_lines, 
                                    output_lines, correctness(expected_output_lines, output_lines))          
                except AssertionError:
                    test_run = TestRun(input_lines, expected_output_lines, [], 0)

                self.test_runs.append(test_run)
            self.score = mean([run.correctness for run in self.test_runs])

            if not cache:
                del self.test_runs

        return self.score

    def save(self, path):
        self.language.copy_source(self.workdir, self.name, path)

    def __del__(self):
        self.language.cleanup(self.workdir, self.name)


import subprocess
import shutil

class Language:
    """
    Language object: describes how programs are compiled and run in a particular language
    """

    def __init__(self, build_cmd, run_cmd, source, artefacts) -> None:
        self.build_cmd = build_cmd
        self.run_cmd = run_cmd
        self.source = source
        self.artefacts = artefacts

    def build(self, workdir, name):
        if self.build_cmd:
            completed_process = subprocess.run(self.build_cmd.format(name=name), capture_output=True, 
                                               cwd=workdir, shell=True)
            return completed_process.stdout, completed_process.stderr
        else:
            return '', ''

    def run(self, workdir, name, input_lines=[]):
        completed_process = subprocess.run(self.run_cmd.format(name=name), 
                                           capture_output=True, 
                                           cwd=workdir, 
                                           input='\n'.join(input_lines).encode(), 
                                           shell=True)
        return completed_process.stdout.decode(), completed_process.stderr.decode()

    def write_source(self, workdir, name, source):
        with open(workdir / self.source.format(name=name), 'w') as f:
            f.write(source)

    def read_source(self, workdir, name):
        with open(workdir / self.source.format(name=name), 'r') as f:
            return f.read()

    def copy_source(self, workdir, name, dest):
        shutil.copy(workdir / self.source.format(name=name), dest)

    def cleanup(self, workdir, name):
        for path in [self.source] + self.artefacts:
            try:
                (workdir / path.format(name=name)).unlink()
            except FileNotFoundError:
                pass

languages = {
    'C++': Language(
        build_cmd='g++ {name}.cpp -o {name}',
        run_cmd='./{name}',
        source='{name}.cpp',
        artefacts=['{name}']
    ),
    'Python': Language(
        build_cmd=None,
        run_cmd='python3 {name}.py',
        source='{name}.py',
        artefacts=[]
    ),
    'Java': Language(
        build_cmd='javac {name}.java',
        run_cmd='java {name}',
        source='{name}.java',
        artefacts=['{name}.class']
    ),
    'Ruby': Language(
        build_cmd=None,
        run_cmd='ruby {name}.rb',
        source='{name}.rb',
        artefacts=[]
    ),
    'Rust': Language(
        build_cmd='rustc {name}.rs',
        run_cmd='./{name}',
        source='{name}.rs',
        artefacts=['{name}']
    ),
    'Go': Language(
        build_cmd=None,
        run_cmd='go run {name}.go',
        source='{name}.go',
        artefacts=[]
    ),
    'Haskell': Language(
        build_cmd=None,
        run_cmd='runhaskell {name}.hs',
        source='{name}.hs',
        artefacts=[]
    ),
    'Scala': Language(
        build_cmd=None,
        run_cmd='scala {name}.scala',
        source='{name}.scala',
        artefacts=[]
    ),
    'Kotlin': Language(
        build_cmd='kotlinc {name}.kt',
        run_cmd='java {name}',
        source='{name}.kt',
        artefacts=['{name}.class']
    ),
    'PHP': Language(
        build_cmd=None,
        run_cmd='php {name}.php',
        source='{name}.php',
        artefacts=[]
    ),
    'C#': Language(
        build_cmd='mcs {name}.cs',
        run_cmd='mono {name}.exe',
        source='{name}.cs',
        artefacts=['{name}.exe']
    ),
    'Swift': Language(
        build_cmd=None,
        run_cmd='swift {name}.swift',
        source='{name}.swift',
        artefacts=[]
    ),
    'D': Language(
        build_cmd=None,
        run_cmd='dmd -of{name} {name}.d',
        source='{name}.d',
        artefacts=['{name}.o']
    ),
    'Julia': Language(
        build_cmd=None,
        run_cmd='julia {name}.jl',
        source='{name}.jl',
        artefacts=[]
    ),
    'Clojure': Language(
        build_cmd=None,
        run_cmd='clj -M {name}.clj',
        source='{name}.clj',
        artefacts=[]
    ),
    'Elixir': Language(
        build_cmd=None,
        run_cmd='elixir {name}.ex',
        source='{name}.ex',
        artefacts=[]
    ),
    'Erlang': Language(
        build_cmd=None,
        run_cmd='erl -noshell -s {name} main -s init stop',
        source='{name}.erl',
        artefacts=[]
    )
}

def language_(language):
    if isinstance(language, Language):
        return language
    else:
        return languages[language]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['programlib']

package_data = \
{'': ['*'], 'programlib': ['programs/*']}

setup_kwargs = {
    'name': 'programlib',
    'version': '3.2.2',
    'description': 'Programs as Objects',
    'long_description': '# Programlib: programs as objects\n\nProgramlib is a tool that turns programs in any programming language into convenient Python objects, letting you run any string as a C++/Python/Clojure/etc program from a Python script.\nThis project is aimed to help develop automatic programming and genetic software improvement systems, though many other applications are possible.\n\n## Installation\n\nProgramlib can be installed with\n\n```\npip install programlib\n```\n\nHowever, you have to also make sure that the programming languages you want to use are installed.\nBy default, programlib uses command line tools that come with the programming languages, i.e. `python3` or `javac`.\n\n## Standard usage\n\nCreate a program object with\n\n```python\nfrom programlib import Program\nprogram = Program(source_code, language=\'C++\')\n```\n\nThis object has\n- a `run` method that runs the program and returns a list of strings it printed to `stdout`. You can optionally provide a list of input strings as well.\n- a `test` method that takes a list of test cases. A test case is a tuple of 2 lists: the first list is the input strings, the second is the expected output strings. The method returns percentage of output strings that matched expectations.\n- a `save` method that will save the source code to a file at the specified path.\n\nSee also `examples`.\n\nCurrently supported programming languages out of the box are C++, Python, Java, Clojure, Ruby, Rust, Go, Haskell, Scala, Kotlin, PHP, C#, Swift, D, Julia, Clojure, Elixir and Erlang.\nSee "Advanced usage" below for instructions on how to add other languages.\n\n## Advanced usage\n\n### Language configuration\n\nWhen you create a program object with a language name like `language=\'C++\'`, `programlib` retrieves an appropriate language configuration from it\'s database.\nIf you have a different opinion on how to compile or run in this language or want to use a language that is not supported out of the box, you can create your own language configuration object:\n\n```python\nfrom programlib import Program, Language\nlanguage = Language(\n        build_cmd=\'g++ {name}.cpp -o {name}\',\n        run_cmd=\'./{name}\',\n        source=\'{name}.cpp\',\n        artefacts=[\'{name}\']\n    )\nprogram = Program(source_code, language=language)\n```\n\n`source` parameter describes the naming convention for the source file (usually `{name}.extension`). Make sure that this parameter contains a `{name}` placeholder, so that `programlib` can keep track of several source files at the same time.\n`build_cmd` and `run_cmd` respectively instruct `programlib` which commands to use to compile and run the program in this language.\n`artefacts` is a list of all the files produced by `build_cmd` command.\nIt is needed to clean up the artefacts when the program object is destroyed.\n\n### Error handling\n\nAny output written to `stderr` is considered an error.\nBy default, any errors at build time or run time will lead to an exception being raised, with 2 exceptions:\n- `test` function that catches exceptions during test cases execution and marks these tests as failed.\n- Setting `program.run(force=True)` or `program.test(force=True)` will make `programlib` ignore all errors.\n\nYou can check `program.stdout` and `program.stderr` to see what the program printed to `stdout` and `stderr` during the last run (or, if in was never run, during build).',
    'author': 'Vadim Liventsev',
    'author_email': 'dev@vadim.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
